require 'test/unit'
require 'tangerine'

# Fake object.
class TestHelper
  def [](x)
    x * 2
  end
  
  def bar
    "BAR"
  end

  def quux
    "QUUUX"
  end

  def hard
    "Th'is ha;rd!"
  end

  def multiline
    "hack\nme!"
  end

  def numbers
    [1, 2, 3, 4, 5]
  end

  def wahr
    true
  end

  def falsch
    false
  end

  def xml
    %{<foo><bar attr="true">text&#x10;'alright'</bar></foo>}
  end
end

# And time stood still...
def Time.now
  Time.at 1095096544
end

class TangerineTest < Test::Unit::TestCase

  # Test verbatim copying
  def test_verbatim
    template = Tangerine.new("foo")
    assert_equal template.expand(nil), "foo"
  end

  def test_simple_replacement
    template = Tangerine.new("foo ,,bar ,,quux quux")
    assert_equal template.expand(TestHelper.new), "foo BAR QUUUX quux"

    template = Tangerine.new("and;now'a,hard$stuff% : ,,hard")
    assert_equal template.expand(TestHelper.new),
                 "and;now'a,hard$stuff% : Th'is ha;rd!"
  end

  def test_eval_replacement
    template = Tangerine.new(
      "foo ,,(bar.downcase) ,,(quux.reverse.tr('U','M')) quux")
    assert_equal template.expand(TestHelper.new), "foo bar XMMMQ quux"

    template = Tangerine.new("1+1 = ,,(1+1)")
    assert_equal template.expand(TestHelper.new), "1+1 = 2"
  end

  def test_multiline
    template = Tangerine.new <<EOF
Multiline test:
,,bar ,,quux
,,multiline
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
Multiline test:
BAR QUUUX
hack
me!
EOF
  end
  
  def test_read_to
    assert_equal "(foo)", Tangerine.new("(foo)").instance_eval {
      @cursor = 0
      @mark = 0
      read_to ?(, ?)
    }
    assert_equal "{fo{o}}", Tangerine.new("{fo{o}}").instance_eval {
      @cursor = 0
      @mark = 0
      read_to ?{, ?}
    }
    assert_raises(RuntimeError, "Hit end of template looking for matching `)'."){
      assert_equal "{fo{o}}", Tangerine.new("{fo{o}}").instance_eval {
        @cursor = 0
        @mark = 0
        read_to ?(, ?)
      }
    }
  end

  def test_comments
    template = Tangerine.new <<EOF
first
,,#comment
last
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
first
last
EOF
  end

  def test_blocks
    template = Tangerine.new <<EOF
first
,,numbers{<,,to_s> }
,,(numbers.reverse){<,,to_s> }
last
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
first
<1> <2> <3> <4> <5> 
<5> <4> <3> <2> <1> 
last
EOF

    template = Tangerine.new <<EOF
first
,,wahr{Yep}
,,falsch{Nope}
last
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
first
Yep

last
EOF
  end
  
  class Blog
    def title
      "chris blogs"
    end
    
    def tagline
      "a blog by christian neukirchen"
    end
    
    def entries
      [] << Entry.new("head", "some <content>") \
         << Entry.new("another", "again")
    end
  end
  
  Entry = Struct.new(:title, :body)
  class Entry    
    # Only entries with a 'd' in the title are considered recent. :P
    def recent?
      title =~ /d/
    end
  end

  def test_blog_example
    template = Tangerine.new <<EOF
<h1>,,title</h1>
<h2>,,tagline</h2>
<p>,,(Time.now)</p>

,,# overview
<ul id="toc">
,,entries{
  [<li>,,(title.upcase)</li>]
}</ul>

,,entries{
<h3>,,title</h3>
  ,,recent?{<em>[New]</em>}
  <p>,,body</p>
<hr />

}
<p>,,title -- ,,tagline. ,,(Time.now.strftime("%d%b%Y").downcase)</p>
EOF
    assert_equal template.expand(Blog.new), <<EOF
<h1>chris blogs</h1>
<h2>a blog by christian neukirchen</h2>
<p>Mon Sep 13 19:29:04 CEST 2004</p>

<ul id="toc">
  [<li>HEAD</li>]
  [<li>ANOTHER</li>]
</ul>

<h3>head</h3>
  <em>[New]</em>
  <p>some <content></p>
<hr />

<h3>another</h3>
  
  <p>again</p>
<hr />


<p>chris blogs -- a blog by christian neukirchen. 13sep2004</p>
EOF
  end

  def test_indent
    template = Tangerine.new <<EOF
first
  ,,|multiline
    ,,|multiline
  * ,,|multiline
  ,,multiline
last
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
first
  hack
  me!
    hack
    me!
  * hack
    me!
  hack
me!
last
EOF
  end

  def test_element_reference
    template = Tangerine.new <<EOF
first
,,[2]
,,["foobar"]
last
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
first
4
foobarfoobar
last
EOF
  end

  def test_xml_escape
    template = Tangerine::XML.new <<EOF
first
,,xml
,,!xml
last
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
first
&lt;foo&gt;&lt;bar attr=&quot;true&quot;&gt;text&amp;#x10;&apos;alright&apos;&lt;/bar&gt;&lt;/foo&gt;
<foo><bar attr="true">text&#x10;\'alright\'</bar></foo>
last
EOF
  end

  class MyTaglib
    def resolve(tag)
      "#{tag}=,,#{tag}"
    end
  end

  def test_taglib
    template = Tangerine.new(",,quux ,,@quux ,,bar ,,@bar", MyTaglib.new)
    assert_equal template.expand(TestHelper.new), "QUUUX quux=QUUUX BAR bar=BAR"

    template = Tangerine.new(",,@quux{,,reverse} ,,@bar", MyTaglib.new)
    assert_equal template.expand(TestHelper.new), "XUUUQ=xuuq bar=BAR"

    assert_raises(RuntimeError, 'Tag not resolvable: foo') {
      template = Tangerine.new(",,@notaglib")
    }
  end

  def test_filelib
    filelib = Tangerine::FileLib.new("test-data", "test-data/subdir")

    assert_raises(RuntimeError, 'Tag not resolvable: foo') {
      template = Tangerine.new(",,@foo", filelib)
    }

    template = Tangerine.new(",,quux ,,@quux ,,bar ,,@bar", filelib)
    assert_equal template.expand(TestHelper.new), "QUUUX Quux! BAR Bar!"
  end

  def test_block_newline
    template = Tangerine.new(",,quux{,,reverse} ,,quux")
    assert_equal template.expand(TestHelper.new), "XUUUQ QUUUX"

    template = Tangerine.new(",,quux{,,reverse}\n,,quux")
    assert_equal template.expand(TestHelper.new), "XUUUQ\nQUUUX"

    template = Tangerine.new(",,quux{,,reverse}  ,,quux")
    assert_equal template.expand(TestHelper.new), "XUUUQ  QUUUX"

    template = Tangerine.new(",,quux{,,reverse}\n\n,,quux")
    assert_equal template.expand(TestHelper.new), "XUUUQ\n\nQUUUX"
  end

  def test_comma_escaping
    template = Tangerine.new("quux")
    assert_equal template.expand(TestHelper.new), "quux"

    template = Tangerine.new(",quux")
    assert_equal template.expand(TestHelper.new), ",quux"

    template = Tangerine.new(",,quux")
    assert_equal template.expand(TestHelper.new), "QUUUX"

    template = Tangerine.new(",,,quux")
    assert_equal template.expand(TestHelper.new), ",,quux"

    template = Tangerine.new(",,,,quux")
    assert_equal template.expand(TestHelper.new), ",,,quux"

    template = Tangerine.new(",,,,,quux")
    assert_equal template.expand(TestHelper.new), ",,,,quux"
  end

  def test_rfc822
    template = Tangerine::RFC822.new <<EOF
Subject: ,,multiline ,,multiline

,,!multiline
EOF
    assert_equal template.expand(TestHelper.new), <<EOF
Subject: hack
 me! hack
 me!

hack
me!
EOF
  end

  def test_errors
    template = Tangerine.new(',,foo')
    assert_raises(NoMethodError, "During template expansion: undefined method `foo' for nil:NilClass") {
      template.expand(nil)
    }
  end
end

